#!/usr/bin/python3
#
# NAME: plot_forecasts.py
#
# PURPOSE:  Quick view of soil moisture forecasts and expected errors for each station
#
# INPUT DATA: name          type      dimension       description
#             ---------------------------------------------------------------------------------
#             yyyymmdd      string    scalar          year,month,day of initial condition of
#                                                     forecast
# OUTPUT:
#
#             ~/LIMBox/plot_forecasts_<yyyymmdd>.png -> Image with time series of
#                                                       station soil moisture forecasts
#
# CREATION HISTORY: Darren Jackson, CIRES/ESRL/PSL, October 2022

import numpy as np 
import os, sys
import pandas as pd
import math

from   datetime import datetime, timedelta
import matplotlib.pyplot as plot
from   matplotlib.ticker import (MultipleLocator, FormatStrFormatter,
                               AutoMinorLocator,MaxNLocator)
import matplotlib.dates as mdates

yyyymmdd = sys.argv[1]

# Collect forecast and expected error data from csv files
fpath = os.path.join(".","data","forecast_data")
f_file = 'SM_'+yyyymmdd+'_forecast.csv'
e_file = 'SM_'+yyyymmdd+'_forecast_error.csv'
if(os.path.exists(fpath+f_file) and os.path.exists(fpath+e_file)):
    sm_fcst = pd.read_csv(fpath+f_file,sep=',')
    sm_err  = pd.read_csv(fpath+e_file,sep=',')

else:
    print('Forecast and/or Expected Error file do not exist for day = '+yyyymmdd)


labels = list(sm_fcst.columns.values)
shape = sm_fcst.shape 
n_stations = shape[1] - 3
ncols = 3
nrows = math.ceil(n_stations/ncols)

# Pull station name from header

fig,axs = plot.subplots(nrows,ncols,figsize=[20,30],facecolor='w',edgecolor='k')
fig.subplots_adjust(hspace= 0.3,wspace=0.3)
station_labels = labels[3:n_stations+3]
ymax = max(sm_fcst.iloc[:,3:].max()+sm_err.iloc[:,3:].max()) + 0.1
init_date = sm_fcst.iloc[0,2]

axs = axs.ravel()

for i in range(n_stations):
#    axs[i].axhline(0.4, color='silver')
    axs[i].plot(sm_fcst.iloc[:,i+3],'blue')
    axs[i].plot(sm_fcst.iloc[:,i+3]+sm_err.iloc[:,i+3],'--',color='blue')
    axs[i].plot(sm_fcst.iloc[:,i+3]-sm_err.iloc[:,i+3],'--',color='blue')
    axs[i].set_ylim([0, ymax])
    axs[i].set_xlim([0, 21])
    axs[i].set_xlabel('Lead Time (Days)',fontsize=16)
    axs[i].set_ylabel('Soil Moisture (m3/m3)',fontsize=16)
    axs[i].tick_params(labelsize=16)
    axs[i].text(1,ymax-0.05,station_labels[i],fontsize=20,weight='bold')
    axs[i].text(1,0.02,"Initialization Date = "+init_date,fontsize=15,weight='bold')
    axs[i].xaxis.set_major_locator(MaxNLocator(7))


# Remove axis for cells with no data
ncell = nrows*ncols
if ncell > n_stations:
    diff = ncell - n_stations
    for n in range(diff):
        axs[n_stations-n+1].axis('off')

# Write plot to image file
fig.savefig("plot_forecasts_"+yyyymmdd+".png")

