#!/usr/bin/python3
#
# NAME: run_forecast.py
#
# PURPOSE: Computes soil moisture forecast from user supplied soil moisture and
#          temperature data
#
# INPUT:   name          type      dimension       description
#          ---------------------------------------------------------------------------------
#          yyyymmdd      string    scalar          year,month,day of initial condition data
#
# INPUT DATA:
#             SM_initial_condition.txt -> soil moisture input file
#             ST_initial_condition.txt -> soil temperature input file
#
# INPUT FORMAT:
#               Date, station 1, station 2, ...., station N   (1st row header,station = name of site)
#               yyyy-mm-dd, value 1, value 2, ..., value N               
#
#               value units are volumetric water content (m3/m3)
#
# OUTPUT DATA:
#
#              SM_<yyyymmdd>_forecast.csv -> SM forecast data from LIM. yyyymmdd = year,month,day
#                                            initial condition
#
# ROUTINES:
#
#              compute_anonaly_scores_for_forecast.py -> Reads soil moisture and temperature initial
#                                                        condition and generates z-score data for LIM model
#              soil_moisture_forecast.py              -> Run LIM with z-score data and generates
#                                                        forecast data
#              convert_forecasted_zscores_to_vwc.py   -> Converts forecasted z-score data back to soil moisture
#                                                        data in volumetric units and creates CSV data file
#                                                        with initial condition and soil moisture forecasts
#                                                        at each station.
#
# NOTES:
#        Use same format for IC data as original soil moisture and
#        and temperature data used to train the LIM.
#
# CREATION HISTORY:  Darren Jackson, CIRES/ESRL/PSL, October 2021

import os,sys

# Read initial condition date
yyyymmdd = sys.argv[1]

os.chdir("./source")

##############################################################################
# Creates initial conditions for soil moisture and temperature data
# for a particular day from which to make a forecast.
#
# Use to create SM_<yyyymmdd>.txt and ST_<yyyymmdd>.txt in format shown above.
#
# NOTE: This code will be unique to each user.
###############################################################################
#os.system("python prepare_data_for_forecast.py "+yyyymmdd)

##################################################################
# These routines make z-scores of initial conditions and compute
# forecast over 21-day period. Output is forecasted soil moisture
# data in volumetric water content units.
##################################################################
os.system("python compute_anomaly_zscores_for_forecast.py "+yyyymmdd)
os.system("python soil_moisture_forecast.py "+yyyymmdd)
os.system("python convert_forecasted_zscores_to_vwc.py "+yyyymmdd)
