#!/usr/bin/python3
#
# NAME: run_training.py
#
# PURPOSE: Generates LIM parameters with a climatology of soil moisture and temperature data
#          that can be used to make forecasts
#
# INPUT DATA:  (Need to reside in ~/LIMBox/data/training_data/)
#             SM_tseries.txt -> soil moisture input file
#             ST_tseries.txt -> soil temperature input file
#
# INPUT FORMAT:
#               Date station1 station2 .... stationN  (1st row header,station = name of site)
#               yyyy-mm-dd  value1 value2 ... valueN  (1st day of time series)            
#               yyyy-mm-dd  value1 value2 ... valueN
#               .
#               .
#               .
#               yyyy-mm-dd  value1 value2 ... valueN (Last day of time series)
#               
#               value units are volumetric water content (m3/m3) for soil moisture and
#               Celsius for soil temperature
#
# ROUTINES:
#
#              compute_mean_annual_cycle.py -> Reads soil moisture and temperature training data
#                                              and generates soil moisture and temperature means
#                                              and annual cycle data
#              compute_anomaly_zscores_residuals.py -> Generates soil moisture and temperature z-score
#                                                      data than can be used to train LIM
#              train_lim.py -> Computes LIM parameters neeed to run the LIM forecast model
#              optimal_structure.py -> Determines soil moisture pattern that can provide most skillful
#                                      forecast
#              error_analysis.py -> Defines which method (Correlation or Projection) has the least amount
#                                   of forecast error for determining Forecasts of Opportunity
#
# NOTES:
#        Use same format for IC data as original soil moisture and
#        and temperature data used to train the LIM.
#
# CREATION HISTORY:  Darren Jackson, CIRES/ESRL/PSL, October 2021

import os

os.chdir("./source")

##############################################################################
# Create time series arrays (time x station) for soil moisture and temperature
# that will be used as input for training code.
#
# Use to create SM_tseries.txt and ST_tseries.txt in format shown above.
#
# NOTE: This code will be unique to each user.
###############################################################################
#os.system("python create_daily_tseries.py 2012 2019")

###############################################################################
# These routines will convert SM and ST time series data into z-scores that are
# used to train the LIM.
###############################################################################
os.system("python compute_mean_annual_cycle.py")
os.system("python compute_anomaly_zscores_residuals.py")

####################################################
# Find LIM parameters that can be used for forecasts
####################################################
os.system("python train_lim.py")
os.system("python optimal_structure.py")
os.system("python error_analysis.py")

