#!/usr/bin/python3
#
# NAME: compute_anomaly_zscores_for_forecast.py
#
# PURPOSE: Reads soil moisture and temperature initial condition (IC)
#          data wanted for a forecast
#          and generates z-scores using climatology derived using
#          compute_anomaly_zscores.py
#
# INPUT DATA:
#              SM(ST)_<yyyymmdd>_ic.txt -> Soil moisture and temperature data for all stations
#                       on specific day that is used as an initial condition for soil moisture
#                       forecast
#              SM(ST)_tseries_mean.txt -> Soil moisture and temperature mean for each station
#              SM(ST)_tseries_ac_noleapday -> Soil moisture and temperature annual cycle for each station
#                       for non-leap years
#              SM(ST)_tseries_ac_leapday -> Soil moisture and temperature annual cycle for each station
#                       for leap years
#              anomaly_mean_std_regr_stats.p -> Mean, standard deviations, regression coefficients
#                       used to compute z-scores from soil moisture and temperature anomaly data
#               
# OUTPUT DATA: 
#              zScores_residST_SM_<yyyyymmdd>.p -> soil moisture anomaly and residual
#                       soil temperature for an initial condition on specified day 
#
# NOTES:
#        Use same format for IC data as original soil moisture and
#        and temperature data used to train the LIM.
#
# CREATION HISTORY:  Darren Jackson, CIRES/ESRL/PSL, October 2021
#----------------------------------------------------------------------

import sys,os
import numpy as np
import pandas as pd
from   numpy import linalg as LA
import pickle
from soil_moisture_lim_mod import tseries_eof_recon


# Read input IC date

yyyymmdd = sys.argv[1]

# Read SM and ST initial condition data

fpath = os.path.join("..","data","forecast_data","initial_conditions")
dpath = os.path.join("..","data","training_data")

fileName_SM = os.path.join(fpath,'SM_'+yyyymmdd+'_ic.txt')
fileName_ST = os.path.join(fpath,'ST_'+yyyymmdd+'_ic.txt')
smdata = pd.read_csv(fileName_SM,sep='\s+') #/s+ is whitespace separater
stdata = pd.read_csv(fileName_ST,sep='\s+') #/s+ is whitespace separater

# Read SM and ST mean and annual cycle data

sm_means = np.loadtxt(os.path.join(dpath,"SM_tseries_mean.txt"))
annCycle_SM_no_leap = np.loadtxt(os.path.join(dpath,"SM_tseries_ac_noleapday.txt"))
annCycle_SM_leap = np.loadtxt(os.path.join(dpath,"SM_tseries_ac_leapday.txt"))

st_means = np.loadtxt(os.path.join(dpath,"ST_tseries_mean.txt"))
annCycle_ST_no_leap = np.loadtxt(os.path.join(dpath,"ST_tseries_ac_noleapday.txt"))
annCycle_ST_leap = np.loadtxt(os.path.join(dpath,"ST_tseries_ac_leapday.txt"))

# Read anomaly means and standard deviation and regression coefficients to compute z-scores

# Read the means and std data used to compute z-scores
f = open(os.path.join(dpath,"anomaly_mean_std_regr_stats.p"),"rb")
allArray_mean = pickle.load(f)
allArray_std = pickle.load(f)
regr = pickle.load(f)
f.close()

# Remove mean from SM and ST IC data

sm_only = smdata.iloc[:,1:].to_numpy()
st_only = stdata.iloc[:,1:].to_numpy()

n_days, n_stations = sm_only.shape
sm_mean_zero = np.empty([n_days,n_stations])
st_mean_zero = np.empty([n_days,n_stations])

for n in np.arange(0, n_days):
    sm_mean_zero[n,:] = sm_only[n,:]-sm_means[:]
    st_mean_zero[n,:] = st_only[n,:]-st_means[:]

# Remove annual cycle from Soil Moisture and Temperature data
# access day of year and leap year information from forecast input file

smdata['doy'] = pd.DatetimeIndex(smdata['Date']).dayofyear
smdata['leap'] = pd.DatetimeIndex(smdata['Date']).is_leap_year

# Define anomaly arrays
n_times, n_stations = sm_only.shape
sm_ac_removed = np.full([n_times,n_stations],np.nan)
st_ac_removed = np.full([n_times,n_stations],np.nan)

for n in np.arange(0, n_times):
    m = smdata['doy'][n] - 1
    if smdata['leap'][n]:
        sm_ac_removed[n,:] = sm_mean_zero[n,:]-annCycle_SM_leap[m,:]
        st_ac_removed[n,:] = st_mean_zero[n,:]-annCycle_ST_leap[m,:]
    else:
        sm_ac_removed[n,:] = sm_mean_zero[n,:]-annCycle_SM_no_leap[m,:]
        st_ac_removed[n,:] = st_mean_zero[n,:]-annCycle_ST_no_leap[m,:]

#Combine the two variables together (time x station/var)

comboArr = np.full([np.shape(np.asarray(sm_ac_removed))[0],np.shape(np.asarray(sm_ac_removed))[1]*2],np.nan)
comboArr[:,0:n_stations]  = st_ac_removed
comboArr[:,n_stations:n_stations*2] = sm_ac_removed

# Put dimensions in right order (station/var x time)

allArray = np.transpose(np.asarray(comboArr))

# Also get the transpose of the combined matrix 

allTranspose = np.transpose(allArray)

# Get the z-scores for each station/variable

#Properly set missing values to NaN from -999 
allArray[allArray<=-999]         = np.nan
allTranspose[allTranspose<=-999] = np.nan
        
# Compute Z-scores
normAll   = np.full([n_stations*2, n_times], np.nan)
normAll_T = np.full([n_times, n_stations*2], np.nan)

for i in np.arange(0,n_stations*2):
    normAll[i,:] = (allArray[i,:] - allArray_mean[i])/allArray_std[i]
    normAll_T[:,i] = (allTranspose[:,i] - allArray_mean[i])/allArray_std[i]

# Create empty arrays to hold T and M as estimated via regression 
# Isolate ST and SM z-scores 
ST = normAll[0:n_stations,:]
SM = normAll[n_stations:n_stations*2,:]

ST_hat = np.zeros([len(regr),len(ST[0,:])])
SM_hat = np.zeros([len(regr),len(SM[0,:])])

# Apply regression
for i in range(len(regr)):
   for j in range(len(ST[0,:])):
       for k in range(len(ST)):
            ST_hat[i,j] += np.nansum(regr[i,k] * SM[k,j])
            SM_hat[i,j] += np.nansum(regr[i,k] * ST[k,j])

# Compute the residual of temperature
#    This removes the part of temperature that is already explained by the moisture field 
resid = ST - ST_hat

# Remove the mean again (center the data)
resid_T = np.transpose(resid)

# Define empty arrays
normResid   = np.full([n_stations, n_times], np.nan)
normResid_T = np.full([n_times, n_stations], np.nan)

# Remove the mean from each station record
if n_times > 1:
    for i in range(n_stations):
        normResid[i,:]   = resid[i,:]   - np.nanmean(resid[i,:])
        normResid_T[:,i] = resid_T[:,i] - np.nanmean(resid_T[:,i])
else:
    normResid[:,:] = resid[:,:]
    normResid_T[:,:] = resid_T[:,:]

# Replace the normalized S.T. with the normalized residual of S.T. in the full combined array 
normAll[0:n_stations,:] = normResid
normAll_T = np.transpose(normAll)

# Fill any missing data with EOF reconstructed data (if needed)
# Save the array with residual S.T. and S.M. z-scores in it 
n_NaNs = np.count_nonzero(np.isnan(normAll))
if n_NaNs > 0:           
    normAll_filled = tseries_eof_recon(normAll,read_eof_data='yes')
    pickle.dump( normAll_filled, open(os.path.join(fpath,"zScores_residST_SM_"+yyyymmdd+".p"), "wb" ) )
else:
    pickle.dump( normAll, open(os.path.join(fpath,"zScores_residST_SM_"+yyyymmdd+".p"), "wb" ) )




