#!/usr/bin/python
#
# NAME: compute_anomaly_zscores_residuals.py
#
# PURPOSE: Reads soil moisture and temperature training data
#          and generates z-scores used for input to determine LIM
#          model parameters
#
# INPUT DATA:
#              SM(ST)_<yyyymmdd>.txt -> Soil moisture and temperature data for all stations
#                       from time series of training data
#                       forecast
#              SM(ST)_tseries_mean.txt -> Soil moisture and temperature mean for each station
#              SM(ST)_tseries_ac_noleapday -> Soil moisture and temperature annual cycle for each station
#                       for non-leap years
#              SM(ST)_tseries_ac_leapday -> Soil moisture and temperature annual cycle for each station
#                       for leap years
# OUTPUT DATA: 
#              zScores_residST_SM.p -> soil moisture anomaly and residual
#                       soil temperature from training data
#              anomaly_mean_std_regr_stats.p -> Mean, standard deviations, regression coefficients
#                       used to compute z-scores from soil moisture and temperature anomaly data
#
# NOTES:
#        Use same format for IC data as original soil moisture and
#        and temperature data used to train the LIM.
#
# CREATION HISTORY:  Darren Jackson, CIRES/ESRL/PSL, October 2021
#----------------------------------------------------------------------

import os
import numpy as np 
from   numpy import linalg as LA
import pandas as pd
from   datetime import datetime
import pickle 
from soil_moisture_lim_mod import tseries_eof_recon

# Read SM and ST station data

dpath=os.path.join("..","data","training_data")
smfile=os.path.join(dpath,"SM_tseries.txt")
stfile=os.path.join(dpath,"ST_tseries.txt")
smdata = pd.read_csv(smfile,sep='\s+')
stdata = pd.read_csv(stfile,sep='\s+')

# Read SM and ST mean and annual cycle data
sm_mnfile=os.path.join(dpath,"SM_tseries_mean.txt")
sm_ac_noleap_file=os.path.join(dpath,"SM_tseries_ac_noleapday.txt")
sm_ac_leap_file=os.path.join(dpath,"SM_tseries_ac_leapday.txt")
sm_means = np.loadtxt(sm_mnfile)
annCycle_SM_no_leap = np.loadtxt(sm_ac_noleap_file)
annCycle_SM_leap = np.loadtxt(sm_ac_leap_file)

st_mnfile=os.path.join(dpath,"ST_tseries_mean.txt")
st_ac_noleap_file=os.path.join(dpath,"ST_tseries_ac_noleapday.txt")
st_ac_leap_file=os.path.join(dpath,"ST_tseries_ac_leapday.txt")
st_means = np.loadtxt(st_mnfile)
annCycle_ST_no_leap = np.loadtxt(st_ac_noleap_file)
annCycle_ST_leap = np.loadtxt(st_ac_leap_file)

# Remove mean from SM and ST data

sm_only = smdata.iloc[:,1:].to_numpy()
st_only = stdata.iloc[:,1:].to_numpy()

n_days, n_stations = sm_only.shape
sm_mean_zero = np.empty([n_days,n_stations])
st_mean_zero = np.empty([n_days,n_stations])

for n in np.arange(0, n_days):
    sm_mean_zero[n,:] = sm_only[n,:]-sm_means[:]
    st_mean_zero[n,:] = st_only[n,:]-st_means[:]

# Remove annual cycle from Soil Moisture and Temperature data

      # access day of year and leap year information
smdata['doy'] = pd.DatetimeIndex(smdata['Date']).dayofyear
smdata['leap'] = pd.DatetimeIndex(smdata['Date']).is_leap_year

      # Define anomaly array
n_times, n_stations = sm_only.shape
sm_ac_removed = np.full([n_times,n_stations],np.nan)
st_ac_removed = np.full([n_times,n_stations],np.nan)

for n in np.arange(0, n_times):
    m = smdata['doy'][n] - 1
    if smdata['leap'][n]:
        sm_ac_removed[n,:] = sm_mean_zero[n,:]-annCycle_SM_leap[m,:]
        st_ac_removed[n,:] = st_mean_zero[n,:]-annCycle_ST_leap[m,:]
    else:
        sm_ac_removed[n,:] = sm_mean_zero[n,:]-annCycle_SM_no_leap[m,:]
        st_ac_removed[n,:] = st_mean_zero[n,:]-annCycle_ST_no_leap[m,:]

#Combine the two variables together (time x station/var)

comboArr = np.full([np.shape(np.asarray(sm_ac_removed))[0],np.shape(np.asarray(sm_ac_removed))[1]*2],np.nan)
comboArr[:,0:n_stations]  = st_ac_removed
comboArr[:,n_stations:n_stations*2] = sm_ac_removed

# Put dimensions in right order (station/var x time)

allArray = np.transpose(np.asarray(comboArr))

# Also get the transpose of the combined matrix 

allTranspose = np.transpose(allArray)

# Get the z-scores for each station/variable

     #Properly set missing values to NaN from -999 
allArray[allArray<=-999]         = np.nan
allTranspose[allTranspose<=-999] = np.nan

     #Define empty arrays
normAll   = np.full([n_stations*2, n_times], np.nan)
normAll_T = np.full([n_times, n_stations*2], np.nan)
allArray_mean = np.full([n_stations*2], np.nan)
allArray_std = np.full([n_stations*2], np.nan)

     #Remove the mean from each station record
for i in np.arange(0,n_stations*2):
    allArray_mean[i] = np.nanmean(allArray[i,:])
    allArray_std[i] = np.nanstd(allArray[i,:])
    normAll[i,:]  = (allArray[i,:]     - allArray_mean[i])/allArray_std[i]
    normAll_T[:,i]= (allTranspose[:,i] - allArray_mean[i])/allArray_std[i]

#Get contemporaneous covariance matrix
nDat = len(normAll)

c0   = np.full([nDat,nDat],np.nan)
for iR in range(nDat):
    for iC in range(nDat): 
        c0[iR,iC] = np.nansum(normAll[iR,:]*normAll_T[:,iC])/np.nansum(np.isfinite(normAll[iR,:]*normAll_T[:,iC]))

# Now we can get the regression coefficients from that matrix, c0 

MM = c0[n_stations:n_stations*2,n_stations:n_stations*2]  #Bottom right quadrant
TM = c0[0:n_stations,n_stations:n_stations*2]   #Top right quadrant 

MM_inv = LA.inv(MM)   #Get the inverse of the moisture covariance matrix  

regr   = np.dot(TM,MM_inv)  #Regression coefficients calculated as product of these two matrices 

# Sum manually in order to do matrix multiplication and ignore missing values 

# Isolate ST and SM z-scores 
ST = normAll[0:n_stations,:]
SM = normAll[n_stations:n_stations*2,:]

# Create empty arrays to hold T and M as estimated via regression 
ST_hat = np.zeros([len(regr),len(ST[0,:])])
SM_hat = np.zeros([len(regr),len(SM[0,:])])

# iterate through rows of X
for i in range(len(regr)):
   # iterate through columns of Y
   for j in range(len(ST[0,:])):
       # iterate through rows of Y
       for k in range(len(ST)):
            ST_hat[i,j] += np.nansum(regr[i,k] * SM[k,j])
            SM_hat[i,j] += np.nansum(regr[i,k] * ST[k,j])

# Compute the residual of temperature
#    This removes the part of temperature that is already explained by the moisture field 
resid = ST - ST_hat

# Remove the mean again (center the data)
resid_T = np.transpose(resid)

# Define empty arrays
normResid   = np.full([n_stations, n_times], np.nan)
normResid_T = np.full([n_times, n_stations], np.nan)

# Remove the mean from each station record
for i in range(n_stations):
    normResid[i,:]   = resid[i,:]   - np.nanmean(resid[i,:])
    normResid_T[:,i] = resid_T[:,i] - np.nanmean(resid_T[:,i])
    
# Replace the normalized S.T. with the normalized residual of S.T. in the full combined array 
normAll[0:n_stations,:] = normResid
normAll_T = np.transpose(normAll)

# Fill any missing data with EOF reconstructed data and
# save the array with residual S.T. and S.M. z-scores in it 
zfile = os.path.join(dpath,"zScores_residST_SM.p")
n_NaNs = np.count_nonzero(np.isnan(normAll))
if n_NaNs > 0:           
    normAll_filled = tseries_eof_recon(normAll)
    pickle.dump( normAll_filled, open( zfile, "wb" ) )
else:
    pickle.dump( normAll, open( zfile, "wb" ) )

# Save the means and std data used to compute z-scores
# Needed to return forecasted z-scores back to volumetric units
an_stats_file = os.path.join(dpath,"anomaly_mean_std_regr_stats.p")
f = open(an_stats_file,"wb")
pickle.dump(allArray_mean, f)
pickle.dump(allArray_std,f)
pickle.dump(regr,f)
f.close()
