#!/usr/bin/python
#
# NAME: prepare_data_for_forecast.py
#
# PURPOSE: Reads available PSL and USCRN station data and constructs ASCII output file of
#          soil moisture and temperature daily data that will be used for forecast
#
# INPUTS:  Name        type     dimension      description
#          ------------------------------------------------------------------------------------------
#          yyyymmdd    string   scalar         year, month, day of soil moisture and temperature data
#                                              to make a foreast

# OUTPUTS:
#          SM_yyyymmdd.txt -> ASCII data file with each data record (or row) containing soil
#                             moisture data for given day at each station.  Data record contains
#                             date, sm station 1, sm station 2, ... , sm station N
#                             date is yyyy-mm-dd format, soil moisture is daily mean in volumetric units.
#          ST_yyyymmdd.txt -> Same as SM data file on soil temperature in units of Celsius.
#
# CALL SEQUENCE:
#                 prepare_data_for_forecast.py <yyyymmdd>
# NOTES:
#         This routine is not a black box routine for users but an example for reading station
#         data from multiple sources and generating an ASCII data file that contains time
#         series of soil moisture and temperature data for multiple stations that is ready to
#         be used by black box routines that compute mean, annual cycle, anomalies,
#         standard deviations, z-scores, etc.
#
# CREATION HISTORY: Darren Jackson, CIRES/ESRL/PSL, October 2021

import sys,os
import numpy as np
from read_data import readSoilData_PSL, readData_CRN, readAllCRN

yyyymmdd = sys.argv[1]
yyyy = yyyymmdd[0:4]
mm   = yyyymmdd[4:6]
dd   = yyyymmdd[6:8]

station_years = [int(yyyy),int(yyyy)]

# Only retrieving one day for now
startDate = yyyy+'-'+mm+'-'+dd
endDate   = yyyy+'-'+mm+'-'+dd

# Read in PSL soil moisture data

dailyAvg_rod = readSoilData_PSL('rod',station_years)
dailyAvg_hbg = readSoilData_PSL('hbg',station_years)
dailyAvg_ptv = readSoilData_PSL('ptv',station_years)
dailyAvg_wls = readSoilData_PSL('wls',station_years)
dailyAvg_lsn = readSoilData_PSL('lsn',station_years)

# Read in the CRN soil moisture data from NSMN web site

# Note that full path may only work for Linux OS
mainPath       = os.path.join("/psd2data","djackson","FACETS","soil_moisture_data")

BodegaFile    = os.path.join(mainPath,"CRNH0203-2018-CA_Bodega_6_WSW.csv")
MercedFile    = os.path.join(mainPath,"CRNH0203-2018-CA_Merced_23_WSW.csv")
ReddingFile   = os.path.join(mainPath,"CRNH0203-2018-CA_Redding_12_WNW.csv")
SBfile            = os.path.join(mainPath,"CRNH0203-2018-CA_Santa_Barbara_11_W.csv")
YosemiteFile  = os.path.join(mainPath,"CRNH0203-2018-CA_Yosemite_Village_12_W.csv")

# Read in CRN data for each station from NSMN data

sm_Bodega    = readData_CRN(BodegaFile)
sm_Merced    = readData_CRN(MercedFile)
sm_Redding   = readData_CRN(ReddingFile)
sm_SantaBarb = readData_CRN(SBfile)
sm_Yosemite  = readData_CRN(YosemiteFile)

# Read in the full CRN files for each station (to grab temperature)

BDG_allCRN = readAllCRN('Bodega_6_WSW',station_years)
MRC_allCRN = readAllCRN('Merced_23_WSW',station_years)
RED_allCRN = readAllCRN('Redding_12_WNW',station_years)
STB_allCRN = readAllCRN('Santa_Barbara_11_W',station_years)
YSM_allCRN = readAllCRN('Yosemite_Village_12_W',station_years)


print('Station      Start Date               End Date   ')
print('-----------------------------------------------------')
print('HBG      ', dailyAvg_hbg.index[0],'   ', dailyAvg_hbg.index[-1])   
print('LSN      ', dailyAvg_lsn.index[0],'   ', dailyAvg_lsn.index[-1])
print('PTV      ', dailyAvg_ptv.index[0],'   ', dailyAvg_ptv.index[-1])
print('ROD      ', dailyAvg_rod.index[0],'   ', dailyAvg_rod.index[-1])
print('WLS      ', dailyAvg_wls.index[0],'   ', dailyAvg_wls.index[-1])
print('BDG      ', sm_Bodega["Date"].iloc[0],'   ', sm_Bodega["Date"].iloc[-1])
print('MRC      ', sm_Merced["Date"].iloc[0],'   ', sm_Merced["Date"].iloc[-1])
print('RED      ', sm_Redding["Date"].iloc[0],'   ', sm_Redding["Date"].iloc[-1])
print('STB      ', sm_SantaBarb["Date"].iloc[0],'   ', sm_SantaBarb["Date"].iloc[-1])
print('YSM      ', sm_Yosemite["Date"].iloc[0],'   ', sm_Yosemite["Date"].iloc[-1])

# Note that using resample to create daily averages results in Date/time index that is used to
# sort data to begin and end for specified time range

HBG_sel = dailyAvg_hbg.loc[(dailyAvg_hbg.index >= startDate) & (dailyAvg_hbg.index <= endDate)]
LSN_sel = dailyAvg_lsn.loc[(dailyAvg_lsn.index >= startDate) & (dailyAvg_lsn.index <= endDate)]
PTV_sel = dailyAvg_ptv.loc[(dailyAvg_ptv.index >= startDate) & (dailyAvg_ptv.index <= endDate)]
ROD_sel = dailyAvg_rod.loc[(dailyAvg_rod.index >= startDate) & (dailyAvg_rod.index <= endDate)]
WLS_sel = dailyAvg_wls.loc[(dailyAvg_wls.index >= startDate) & (dailyAvg_wls.index <= endDate)]

# Select time period in soil moisture CRN data 
BDG_sel = sm_Bodega.loc[((sm_Bodega.Date       >= startDate) & (sm_Bodega.Date <= endDate))]
MRC_sel = sm_Merced.loc[((sm_Merced.Date       >= startDate) & (sm_Merced.Date <= endDate))] 
RED_sel = sm_Redding.loc[((sm_Redding.Date     >= startDate) & (sm_Redding.Date <= endDate))]
STB_sel = sm_SantaBarb.loc[((sm_SantaBarb.Date >= startDate) & (sm_SantaBarb.Date <= endDate))]
YSM_sel = sm_Yosemite.loc[((sm_Yosemite.Date   >= startDate) & (sm_Yosemite.Date <= endDate))]

# Select the same time period for everything in soil temperature arrays as well 
BDG_AllSel = BDG_allCRN.loc[((BDG_allCRN.Date   >= startDate) & (BDG_allCRN.Date <= endDate))]
MRC_AllSel = MRC_allCRN.loc[((MRC_allCRN.Date   >= startDate) & (MRC_allCRN.Date <= endDate))] 
RED_AllSel = RED_allCRN.loc[((RED_allCRN.Date   >= startDate) & (RED_allCRN.Date <= endDate))]
STB_AllSel = STB_allCRN.loc[((STB_allCRN.Date   >= startDate) & (STB_allCRN.Date <= endDate))]
YSM_AllSel = YSM_allCRN.loc[((YSM_allCRN.Date   >= startDate) & (YSM_allCRN.Date <= endDate))]

#Handle missing data
BDG_AllSel.loc[(BDG_AllSel.ST_10_Daily<=-99.0), :] = np.nan
MRC_AllSel.loc[(MRC_AllSel.ST_10_Daily<=-99.0), :] = np.nan
RED_AllSel.loc[(RED_AllSel.ST_10_Daily<=-99.0), :] = np.nan
STB_AllSel.loc[(STB_AllSel.ST_10_Daily<=-99.0), :] = np.nan
YSM_AllSel.loc[(YSM_AllSel.ST_10_Daily<=-99.0), :] = np.nan

BDG_AllSel.loc[(BDG_AllSel.SM_10_Daily<=-99.0), :] = np.nan
MRC_AllSel.loc[(MRC_AllSel.SM_10_Daily<=-99.0), :] = np.nan
RED_AllSel.loc[(RED_AllSel.SM_10_Daily<=-99.0), :] = np.nan
STB_AllSel.loc[(STB_AllSel.SM_10_Daily<=-99.0), :] = np.nan
YSM_AllSel.loc[(YSM_AllSel.SM_10_Daily<=-99.0), :] = np.nan

#Define date array
dates      = BDG_sel.Date

# Write daily mean soil moisture and temperature time series for all stations

fileName_SM = '../data/forecast_data/initial_conditions/SM_'+yyyymmdd+'_ic.txt'
fileName_ST = '../data/forecast_data/initial_conditions/ST_'+yyyymmdd+'_ic.txt'

with open(fileName_SM,'w+') as f:
    f.write("%s %s %s %s %s %s %s %s %s %s %s \n" % ('Date','HBG_sm_10cm','LSN_sm_10cm',
                                                     'PTV_sm_10cm','ROD_sm_10cm','WLS_sm_10cm',
                                                     'BDG_sm_10cm','MRC_sm_10cm','RED_sm_10cm',
                                                     'STB_sm_10cm','YSM_sm_10cm'))
    for n in range(len(dates)):
        f.write("%s %f %f %f %f %f %f %f %f %f %f \n" % (str(np.asarray(dates)[n])[0:10],
                                                             np.asarray(HBG_sel.soilM_10)[n],
                                                             np.asarray(LSN_sel.soilM_10)[n],
                                                             np.asarray(PTV_sel.soilM_10)[n],
                                                             np.asarray(ROD_sel.soilM_10)[n],
                                                             np.asarray(WLS_sel.soilM_10)[n],
# SM data not QC'd by NSMN
#                                                             np.asarray(BDG_AllSel.SM_10_Daily)[n],
#                                                             np.asarray(MRC_AllSel.SM_10_Daily)[n],
#                                                             np.asarray(RED_AllSel.SM_10_Daily)[n],
#                                                             np.asarray(STB_AllSel.SM_10_Daily)[n],
#                                                             np.asarray(YSM_AllSel.SM_10_Daily)[n]))
                                                             np.asarray(BDG_sel.sm_10cm)[n],
                                                             np.asarray(MRC_sel.sm_10cm)[n],
                                                             np.asarray(RED_sel.sm_10cm)[n],
                                                             np.asarray(STB_sel.sm_10cm)[n],
                                                             np.asarray(YSM_sel.sm_10cm)[n]))


with open(fileName_ST,'w+') as f:
    f.write("%s %s %s %s %s %s %s %s %s %s %s \n" % ('Date','HBG_st_10cm','LSN_st_10cm',
                                                     'PTV_st_10cm','ROD_st_10cm','WLS_st_10cm',
                                                     'BDG_st_10cm','MRC_st_10cm','RED_st_10cm',
                                                     'STB_st_10cm','YSM_st_10cm'))
    for n in range(len(dates)):
        f.write("%s %f %f %f %f %f %f %f %f %f %f \n" % (str(np.asarray(dates)[n])[0:10],
                                                             np.asarray(HBG_sel.soilT_10)[n],
                                                             np.asarray(LSN_sel.soilT_10)[n],
                                                             np.asarray(PTV_sel.soilT_10)[n],
                                                             np.asarray(ROD_sel.soilT_10)[n],
                                                             np.asarray(WLS_sel.soilT_10)[n],
                                                             np.asarray(BDG_AllSel.ST_10_Daily)[n],
                                                             np.asarray(MRC_AllSel.ST_10_Daily)[n],
                                                             np.asarray(RED_AllSel.ST_10_Daily)[n],
                                                             np.asarray(STB_AllSel.ST_10_Daily)[n],
                                                             np.asarray(YSM_AllSel.ST_10_Daily)[n]))


