#!/usr/bin/python
#
# NAME: read_data.py
#
# PURPOSE: Module containing three functions used to read PSL and USCRN soil moisture
#          and temperature data
#
# NOTES:  This code is not a black box for users but a module containing functions to read the
#         PSL and USCRN data.  These functions are to be called by create_daily_tseries.py
#         which is used to create the daily mean SM and ST time series that are to be
#         used for computing mean, annual cycle, and anomaly data
#
# CREATION HISTORY: Darren Jackson, CIRES/ESRL/PSL, September 2021

import os
import numpy as np 
import pandas as pd
from   datetime import datetime
from   soil_moisture_lim_mod import getLocTime

def readSoilData_PSL(station,stationYears):

# 
# PURPOSE: Reads PSL hourly soil moisture and temperature data that have been previously
#          QC'd and saved in ASCII data format and return daily averaged data
#
# INPUTS:  name          type     dimension   description
#          ---------------------------------------------------------------------------------
#          station       string   scalar      station 3 letter name (ex. Rio Nido = rod)
#          stationYears  integer  list        list of years (ex. [2011,2012])
# OUTPUTS:
#          dailyAvg      float    
# NOTES:
#          Data paths hardwired into this code


# Note: only works for Linux OS
    mainPath = os.path.join("/psd2data","Soil_Moisture_qc1","final_data")
    
    #Initialize arrays to save all the data at this station
    appended_data = []

    #Loop over years in selection 
    for iYr in np.arange(0,len(stationYears)):
        colNames = ["year","doy","time","sfcT","RH","P",
                    "soilT_10","soilT_15","soilM_10","soilM_15"]
        fileName = os.path.join(mainPath,station,"hourly",station+'_'+str(stationYears[iYr])+'_hr.dat')
        
        #Read in file as a pandas dataframe
        data         = pd.read_csv(fileName, sep=",", header=None, usecols=(np.arange(1,len(colNames)+1)))
        data.columns = colNames
        
        # Get local time instead of UTC
        localDate = getLocTime(data.doy,data.year,data.time*100)

        # Add local time as column to pandas DF
        data.insert(0,"LocalTime",localDate,True)
        
        #Append this year's data onto the end of the last year
        if iYr==0:
            appended_data = data
        elif iYr>0:
            appended_data = pd.concat([appended_data, data],sort=False, ignore_index=True)
            
        # Remove any missing values (typically stored as -1000 or -999,
        # but if SM<0, you know it's wrong [Bob's method])
        appended_data.loc[(appended_data.soilM_10<0), :] = np.nan
        appended_data.loc[(appended_data.soilT_10<-50), :] = np.nan

    # Compute daily average from hourly data
    dailyAvg = appended_data.resample('D', on = 'LocalTime').mean() 

    # Return daily averaged data
    return dailyAvg

def readData_CRN(fileName):

# 
# PURPOSE: Reads USCRN soil moisture data from NSMN web site which has applied additional QC
#          above and beyond USCRN efforts.
#
# INPUTS:  name          type     dimension   description
#          ---------------------------------------------------------------------------------
#          FileName      string   scalar      Full path input file name
# OUTPUTS:
#          smData        float    Pandas DF   Soil moisture output file



    # Read in file data
    smData         = pd.read_csv(fileName, sep=",", header=0, skiprows=1)
    smData.columns = ["network","stationID","Year","Month","Day","Doy","sm_5cm","sm_10cm",
                      "sm_20cm","sm_50cm","sm_100cm","Lat","Lon","flag_depth1","flag_depth2",
                      "flag_depth3","flag_depth4","flag_depth5"]
    
    # Get date as a single value
    dt = np.empty(len(smData.Year), dtype='datetime64[s]')
    for iT in range(len(smData.Year)):
        dt[iT] = datetime(year=int(smData.Year[iT]), month=int(smData.Month[iT]), day=int(smData.Day[iT]))

    # Add local time as column to pandas DF and drop year/month/day
    smData.insert(0,"Date",dt,True)
    smData = smData.drop('Year', axis=1)
    smData = smData.drop('Month', axis=1)
    smData = smData.drop('Day', axis=1)
    
    return smData

def readAllCRN(stationSpec,stationYears):

# 
# PURPOSE: Reads USCRN soil moisture and temperature data. Used to get USCRN soil temperature data.
#
# INPUTS:  name          type     dimension   description
#          ---------------------------------------------------------------------------------
#          FileName      string   scalar      Full path input file name
#          stationYears  integer  list        list of years (ex. [2011,2012])
# OUTPUTS:
#          crn_data      float    2D Data     CRN Soil moisture and temperature data, Pandas DF


#    mainPath = '/psd2data/djackson/FACETS/soil_moisture_data/CRND0103-'
    mainPath = os.path.join("/psd2data","djackson","FACETS","soil_moisture_data")
    
    crn_data = []
    for n in np.arange(0,len(stationYears)):
        iYr = stationYears[n]
        fileName = mainPath+'/CRND0103-'+str(iYr)+'-CA_'+stationSpec+'.txt'
        yrData   = pd.read_csv(fileName,sep='\s+',header=None) #/s+ is whitespace separater
        yrData.columns = ["WBANNO","Date","CRX_VN","Lon","Lat","T_Daily_Max","T_Daily_Min","T_Daily_Mean",
                         "T_Daily_Avg","P_Daily_Calc","SolaRad_Daily","Sfc_T_Daily_Type","Sfc_T_Daily_Max",
                         "Sfc_T_Daily_Min","Sfc_T_Daily_Avg","RH_Daily_Max","RH_Daily_Min","RH_Daily_Avg",
                         "SM_5_Daily","SM_10_Daily","SM_20_Daily","SM_50_Daily","SM_100_Daily",
                         "ST_5_Daily", "ST_10_Daily","ST_20_Daily","ST_50_Daily","ST_100_Daily"]
        if n==0:
            crn_data = yrData
        else:
            crn_data = pd.concat([crn_data, yrData],sort=False, ignore_index=True)
            
    #Put date in datetime form
    crn_data.Date = pd.to_datetime(crn_data.Date,format = "%Y%m%d")

    return crn_data








