#!/usr/bin/python3
#
# NAME: soil_moisture_forecast.py
#
# PURPOSE: Reads soil moisture and temperature initial condition (IC)
#          z-scores derived from compute_anomaly_zscores_for_forecast.py
#          and generates a forecast from the IC data and determines if the
#          forecast is a forecast of opportunity.
#
# INPUT:   name          type      dimension       description
#          ---------------------------------------------------------------
#          yyyymmdd      string    scalar          year,month,day
#
# INPUT DATA:
#
#          SM_initial_condition.txt     -> Reads soil moisture input file to retrieve date information
#          zscore_residST_SM_forecast.p -> initial condition data from which to make forecast
#          optimal_structure_peak.p     -> optimal structure needed to assess forecast of opportunity (FoO)
#          lim_forecast_matrices.p      -> LIM parameters computed from training data needed for forecast
#          optimal_projection_pcorr.p   -> Tercile and histogram data computed from correlation data for FoO
#
# OUTPUT DATA:
#
#          forecasted_SM_zscores.p      -> SM forecast data from LIM 
#
# NOTES:
#        Use same format for IC data as original soil moisture and
#        and temperature data used to train the LIM.
#
# CREATION HISTORY:  Darren Jackson, CIRES/ESRL/PSL, October 2021
#----------------------------------------------------------------------

import numpy as np 
import os, sys
import pandas as pd
import pickle

#-----------------
# User input
#-----------------
yyyymmdd = sys.argv[1]
#method_type = sys.argv[2]

#------------------------
# Read Input Data
#------------------------

opath = os.path.join("..","data","forecast_data")
fpath = os.path.join(opath,"initial_conditions")
dpath = os.path.join("..","data","lim_parameter_data")

# Input initial condition data to retrieve dates
fileName_SM = os.path.join(fpath,'SM_'+yyyymmdd+'_ic.txt')
smdata = pd.read_csv(fileName_SM,sep='\s+')  #/s+ is whitespace separater

# Input zScores from forecast data
filePath_zscore2 = os.path.join(fpath,'zScores_residST_SM_'+yyyymmdd+'.p')
if os.path.isfile(filePath_zscore2)==True:
    fdata = pickle.load( open( filePath_zscore2, "rb" ) )
else: 
    sys.exit('   WARNING: Forecast Zscore data file does not exist! Cannot continue.')   
    
filePath_structure = os.path.join(dpath,'optimal_structure_peak.p')
if os.path.isfile(filePath_structure)==True:
    f = open(filePath_structure,"rb")
    Tau_peak = pickle.load(f)
    gamma_all = pickle.load(f)
    optimStruct = pickle.load(f)
    f.close()
else:
     sys.exit('   WARNING: Optimal structure data file does not exist! Cannot continue.')

filePath_forecast_matrices =os.path.join(dpath,'lim_forecast_matrices.p')
if os.path.isfile(filePath_forecast_matrices)==True:
    f = open(filePath_forecast_matrices,"rb")
    tau0 = pickle.load(f)
    G_tau_3d = pickle.load(f)
    Exp_stationErr = pickle.load(f)
    f.close()
else: 
    sys.exit('   WARNING: forecast matrix data file does not exist! Cannot continue.')
    
#filePath_projection = dpath+'optimal_projection.p'
#if os.path.isfile(filePath_projection)==True:
#    f = open(filePath_projection,"rb")
#    projTrcile = pickle.load(f)
#    optProj = pickle.load(f)
#    f.close()
#else:
#    sys.exit('   WARNING: Projection histogram data file does not exist! Cannot continue.')

filePath_pcorr = os.path.join(dpath,'optimal_projection_pcorr.p')
if os.path.isfile(filePath_pcorr)==True:
    f = open(filePath_pcorr,"rb")
    pcorr_Trcile = pickle.load(f)
    pcorr = pickle.load(f)
    f.close()
else:
    sys.exit('   WARNING: Correlation data file does not exist! Cannot continue.')
    
#print('Lag time in days of best forecast skill: ',tau0)
#print('Lag time in days of maximum amplification: ',Tau_peak)

#------------------------------------------
# Define parameters need to assign arrays
#------------------------------------------

# Define size of forecast data array 
sizes = np.shape(fdata)
nFDat  = sizes[0]
nFT    = sizes[1]
nFDat2 = int(nFDat/2)

# Number of lags determined in train_lim.ipynb G_tau output data
Gsize = np.shape(G_tau_3d)
n_lags = Gsize[0]


#---------------
# Make forecast
#---------------
  
#Define empty array to store forecast at each time step and each lag
fcst     = np.full([nFDat,nFT,n_lags], np.nan)
    
for ilag in range(1,n_lags):   #Loop over number of lags to be computed. Error at ilag=0 set to zero later.

    #Define Green function at this lag using input data
    G_tau = G_tau_3d[ilag,:,:]
       
    for iT in range(nFT):    #Loop over timesteps 
        
        # --- Get forecast data --- #
        #If there is any missing data at this time...
        if ( np.isnan(fdata[:,iT]).any() ): 
            
            #Indices where NaNs are *not*
            iNan = np.where(~np.isnan(fdata[:,iT]))
            
            #Make replica matrix replacing NaNs with zeros so that it won't count for other stations (but is computed)
            replZeros          = np.zeros(np.shape(fdata[:,iT]))
            replZeros[iNan[0]] = fdata[iNan[0],iT]
            
            #Get forecast using zeros array so that stations w/o NaN are still computed properly 
            fcst[:,iT,ilag] = np.dot(G_tau, replZeros)
            
            #Replacing stations that had missing data with a NaN (no forecast if no data)
            fcst[np.where(np.isnan(fdata[:,iT])),iT,ilag] = np.nan
        else:     
            #If no data is missing, just compute forecast as normal - G(Tau) * X(:,t)
            fcst[:,iT,ilag] = np.dot(G_tau, fdata[:,iT])         
    
# Set first forecast value at ilag=0 to the initial condition: 
fcst[:,:,0]=fdata[:,:]

#---------------------------------------------------------------------------
# Two methods to compute histograms used to determine cut-off of top tercile
#   Only using correlation method for now
#---------------------------------------------------------------------------

cutoff_F   = np.full([nFT], np.nan)

#if (method_type == 'P'):
## Projection method
#    for iT in range(nFT):
#    #For matrix multiplication with nan, take nansum after *
#        cutoff_F[iT]   = np.nansum(np.transpose(optimStruct) * data[:,iT])  
#        if (np.abs(cutoff_F[iT]) > projTrcile):
#            print('Forecast of Opportunity identified for day = ',smdata['Date'][iT])
#        else:
#            print('No Forecast of Opportunity found for day = ',smdata['Date'][iT])
#else:

# Correlation method
valid = np.full([nFT], True)
valid_stations = np.full([nFT], np.nan)
for iT in range(nFT):
    valid = np.where(np.isfinite(fdata[:,iT]))[0]
    valid_stations[iT] = len(valid)
    if len(valid) > nFDat/2:
        R = np.corrcoef(optimStruct[valid],fdata[valid,iT])
        cutoff_F[iT] = R[0,1]
        
if (np.abs(cutoff_F[iT]) > pcorr_Trcile):
    print('Forecast of Opportunity identified for day = ',smdata['Date'][iT])
    FoO = 1
else:
    print('No Forecast of Opportunity found for day = ',smdata['Date'][iT])
    FoO = 0

#-------------------------------------------------------------
# Write forecasted zscores and expected errors to output file
#-------------------------------------------------------------

f = open(os.path.join(opath,"Zscores_"+yyyymmdd+"_forecast.p"),"wb")
pickle.dump(FoO,f)
pickle.dump(cutoff_F[iT],f)
pickle.dump(fcst, f)
pickle.dump(Exp_stationErr, f)
f.close()

