#!/usr/bin/python3
#
# NAME: soil_moisture_lim_mod.py
#
# PURPOSE: Module containing functions used to
#     1. test_q_eigenvalues: test eigenvalues of Q for positivity
#     2. getLocTime:         output local time and date
#     3. leapyear:           identify leap year
#     4. tseries_eof_recon:  fill missing data in soil moisture/temperature training data &
#                            fill missing data in forecast data input
#     5. clean_data_dir:     NOT AVAILABLE, removes all data files from specified data directories
#
# CREATION HISTORY: Darren Jackson, CIRES/ESRL/PSL, October 2021

import sys,os
import pandas as pd
import numpy as np 
from   numpy import linalg as LA
import scipy 
from   scipy import stats
import pickle
from   datetime import datetime
from   dateutil import tz

# Define a function that determines if eigenvalues of Q are all positive

def test_q_eigenvalues(tau0, data):

# Define size of data array 
    sizes = np.shape(data)
    nDat  = sizes[0]
    nT    = sizes[1]
    
#Define transpose of data 
    data_T = np.transpose(data)

#Get contemporaneous covariance matrix (c0)
    c0   = np.full([nDat,nDat],np.nan)
    for iR in range(nDat):
        for iC in range(nDat): 
            c0[iR,iC] = np.nansum(data[iR,:]*data_T[:,iC])/np.nansum(np.isfinite(data[iR,:]*data_T[:,iC]))

#Get lagged covariance matrix... (cT)
#   First define X(t + Tau)
    xLagged = np.full([nDat,nT-tau0],np.nan)
    for iT in range(nT-tau0):
        xLagged[:,iT] = data[:,iT+tau0]        

#   Next compute the actual lag covariance matrix, where missing data is a concern
    cT = np.full([nDat,nDat],np.nan)
    for iR in range(nDat):
        for iC in range(nDat):
            cT[iR,iC] = np.nansum(xLagged[iR,:]*data_T[:-tau0,iC])/np.nansum(np.isfinite((xLagged[iR,:]*data_T[:-tau0,iC])))

# The Green function is defined as the product of the lagged and inverse contemporaneous covariance matrices 
    G = cT.dot(LA.inv(c0)) 

# Eigenanalysis of G 
    g,u = LA.eig(G)

    iSort = g.argsort()[::-1]    #Sort the eigen values and vectors in order
    g     = g[iSort]
    u     = u[:,iSort]

# Eigenanalysis of G^T ( define adjoints (v) based on transpose of G(tau) )
    eigVal_T, v = LA.eig(np.transpose(G))

    iSortT      = eigVal_T.argsort()[::-1]
    eigVal_T    = eigVal_T[iSortT]
    v           = v[:,iSortT]
    
# But modes should ultimately be sorted by decreasing decay time (i.e., decreasing values of 1/beta.real)
 
# Compute Beta  
    b_tau   = np.log(g)
    b_alpha = b_tau/tau0

# Sort data by decreasing decay time 
    sortVal = -1/b_alpha.real              #Decay time 

    iSort2 = sortVal.argsort()[::-1]      #Sorted indices 
    u      = u[:,iSort2]
    v      = v[:,iSort2]
    g      = g[iSort2]
    b_alpha = b_alpha[iSort2]

# Make diagonal array of Beta (values should be negative)
    beta = np.zeros((nDat, nDat), complex)
    np.fill_diagonal(beta, b_alpha)

# Normalize u 
    normFactors  = np.dot(np.transpose(u),v)
    normU        = np.dot(u,LA.inv(normFactors))

# Redefine 'u' as normU to avoid any confusion 
    u = normU

# --- Compute L matrix as u*beta*v_transpose --- #
    L     = np.dot(u,np.dot(beta,np.transpose(v)))

# --- Define the Q matrix --- # 
    Q_minus = np.dot(L,c0) + np.dot(c0,np.transpose(L))
    Q = -Q_minus

# Test that all eigenvalues of Q are positive
# Eigenanalysis of G 
    Q_eigval,Q_eigvec = LA.eig(Q)

#Sort the eigenvalues in descending order 
    iSortQ   = Q_eigval.argsort()[::-1]
    Q_eigval = Q_eigval[iSortQ]
    Q_eigvec = Q_eigvec[:,iSortQ]

    if np.any(Q_eigval<0)==True:
        Q_eigval_all_positive = False
    else:
        Q_eigval_all_positive = True
        
    return Q_eigval_all_positive 

# Output: Local Date and Time array in 

def getLocTime(calenDay,year,time):
    localDate  = np.empty(len(time), dtype='datetime64[s]')     #Empty array to save local time into 
    
    # Settings for time conversion 
    from_zone = tz.gettz('UTC')                     #Time zone data is currently saved in
    to_zone   = tz.gettz('America/Los_Angeles')     #Time zone you *want* data to be in instead 
    
    #-- Convert UTC to Pacific Time --#
    for iT in np.arange(0,len(time)):
        strDay  = str(int(np.asarray(calenDay[iT]))).zfill(3)
        strYr   = str(int(np.asarray(year[iT])))
        strHr   = str(int(np.asarray(time[iT]))).zfill(4)[:2]
        
        utc     = datetime.strptime(strYr+' '+strDay+' '+strHr, '%Y %j %H')
        # Tell the datetime object that it's in UTC time zone  
        utc     = utc.replace(tzinfo=from_zone)
        # Convert time zone
        pacTime = utc.astimezone(to_zone)
        
        # Save local date for given time 
        localDate[iT] = pacTime.replace(tzinfo=None)
    # ------- 
    
    return localDate


def leapyear(year):
    leap=0
    if year%100==0:
        if year%400==0:
            leap=1
    else:
        if year%4==0:
            leap=1
    return leap


# This function uses EOFs to fill missing data in the soil moisture and temperature training data
# and outputs the EOFs to file so initial condition data used for forecasts that might contain
# missing data can be filled using EOFs

# Input and output data dimensions are location x time

# Input missing data with input needs to be set to nan

# Output fills missing data in input data with EOF reconstructed data

# Using optional 2nd argument "eof_only" will return output of full EOF reconstructed time series

def tseries_eof_recon(idata,read_eof_data=None,eof_tseries=None):

    # Realign input data so it is location x time
    zdata = np.copy(idata)

    # Define sizes and transpose array
    sizes = np.shape(zdata)
    nDat = sizes[0]
    nT = sizes[1]
    c0 = np.full([nDat, nDat],np.nan)  # initialize covariance matrix
    zdata_T = np.transpose(zdata)
    
    # Check if EOF data path specified to retrieve data
    if read_eof_data:
        dpath = '../data/training_data'
        filePath_eof = dpath+'/zscore_eof_data.p'
        if os.path.isfile(filePath_eof)==True:
            f = open(filePath_eof,"rb")
            g = pickle.load(f)
            u = pickle.load(f)
            f.close()
        else:
            sys.exit('   WARNING: Projection eof data file does not exist! Cannot continue.')
        
    else:
        for iR in range(nDat):      # Compute eigenvectors from input data
            for iC in range(nDat):
                # Contemporaneous covariance matrix:
                c0[iR,iC] = np.nansum(zdata[iR,:]*zdata_T[:,iC]) / np.nansum(np.isfinite(zdata[iR,:]*zdata_T[:,iC]))
    
        g, u = LA.eig(c0) # Compute eigenvectors or modes(u) and eigenvalues(g) of c0
        iSort = g.argsort()[::-1]    # Sort the eigen values and vectors in order 
        g     = g[iSort]             # e-values
        u     = u[:,iSort]           # e-vectors

    sizes = np.shape(g)
    nEof = sizes[0]
    
    # Compute the PCs
    pc = np.full([nT,nDat],np.nan)
    for iT in range(nT):
        valid = np.where(np.isfinite(zdata[:,iT]))
        sz = np.shape(valid)
        if sz[1] > 0:
            for j in range(nEof):
                # nan accounted for in zdata sum and use valid index for denominator sum
                pc[iT,j] = np.nansum(zdata[:,iT]*u[:,j])/np.sum(u[valid,j]**2) 

        else:
            print("Missing data at all locations for time = ",iT)
            print("Can not fill time series")
            sys.exit()
    
    # Reconstruct data using PCs
    zdata_eof = np.full([nT,nDat],np.nan)
    for iT in range(nT):
        for i in range(nDat):
            zdata_eof[iT,i] = np.sum(pc[iT,:]*u[i,:])

    # Output eigenvalues and eigenvectors to file
    opath = '../data/training_data/'
    f = open(opath+"zscore_eof_data.p","wb")
    pickle.dump(g, f)
    pickle.dump(u, f)
    f.close()

    # Fill missing data in original data and return new time series
    zdata_eof_T = np.transpose(zdata_eof)
    if eof_tseries:
        return zdata_eof_T
    else:
        iNan = np.where(np.isnan(zdata))
        zdata_fill = np.copy(zdata)
        zdata_fill[iNan] = zdata_eof_T[iNan]        
        return zdata_fill

