#!/usr/bin/python3
#
# NAME: train_lim.py
#
# PURPOSE: Determines the LIM parameters needed to the make best soil moisture forecast
#
# INPUT FILES:  Name                  Type     Dimension          Origin Code                           
#               ------------------------------------------------------------------------------------------
#               zScores_residST_SM.p  Pickle   [stations x time]  compute_anomaly_zscores_residuals.py  
#
# OUTPUT FILES:
#               lim_forecast_matrices.p -> Contains the lag time with the lowest error for the 7-14 day
#                                          forecasts and the associated LIM parameters needed to run
#                                          the model forecasts
# NOTES:
#         Linear inverse model is built around an array of data supplied by the user <b>(as a pickle file)</b>.
#         Dimensions should be in the form of station by time. (Example, [10 x 2049] in the case of the
#         soil moisture data analyzed in Penland et al. 2021). It is assumed that the data has already been
#         centered (i.e., has zero mean) and that the annual cycle has been removed by using codes
#         compute_mean_annual_cycle.py and compute_anomaly_zscores_residuals.py.  Input lag (Tau_0) of 7 was
#         used in manuscript.
#
# CREATION HISTORY:  Meg D. Fowler, CIRES/PSL, February 2020
#                    Modified by Darren Jackson, CIRES/PSL October 2021
#                    1/25/22 - Found bug with lag=0 G_tau data and corrected zero power matrix calculation
# 

#------------------------
# Import Libraries
#------------------------

import numpy as np 
from   numpy import linalg as LA
from   numpy.linalg import matrix_power

import scipy 
from   scipy import stats 
import pandas as pd
import pickle
import os, sys

from soil_moisture_lim_mod import test_q_eigenvalues
from LIM_analysis import LIM

#--------------------------------------------------
# Input soil moisture and temperature anomaly data 
#--------------------------------------------------

ipath = os.path.join( "..","data","training_data")
opath = os.path.join("..","data","lim_parameter_data")
filePath_zscores  = os.path.join(ipath,"zScores_residST_SM.p")

if os.path.isfile(filePath_zscores)==True:
    data = pickle.load( open( filePath_zscores, "rb" ) )
else: 
   sys.exit('File = '+filePath_zscores+' does not exit. Program terminated.')

sizes = np.shape(data)
nDat = sizes[0]
nT = sizes[1]

print('Train LIM with ',nDat,' moisture and temperature data locations and ',nT,' time  points') 

#-----------------------------------------
# Tau test and Nyquist Mode check
#-----------------------------------------

# Define a series of different values of tau0 to test over 

lags = np.arange(5,15,1) # Test lags from 5 to 15

# Check each lag in lags array to see if/where a Nyquist mode arises

valid_tau0 = False
for iLag in range(len(lags)):
    #Compute LIM at this lag
    b_alpha_fnct,L_fnct,Q_fnct,G_fnct,c0_fnct,cT_fnct,u_fnct,v_fnct,g_fnct,periods_fnct,decayT_fnct = LIM(data,lags[iLag])  
    nyq_check = b_alpha_fnct.imag*lags[iLag]    #Make array of beta_imaginary * lags, used to check for Nyquist mode 
    
    if np.any((nyq_check>=3.141) & (nyq_check<=3.142))==False: 
        if (not valid_tau0):
            valid_tau0 = True
            tau0_arr = np.asarray(lags[iLag])
        else:
            tau0_arr = np.append(tau0_arr,lags[iLag])

try:
    tau0_arr[0]
except:
    print("No Lags passed the Nyquist mode test")
    print("Do not write LIM parameters and exit program")
    sys.exit("1")
else:
    print("Lag times, in days, to test for hindcast errors: ",tau0_arr)
    
#------------------------------------------------------------------------------
# Compute error of hindcasts generated from tau0 values passing Tau test
#------------------------------------------------------------------------------

#   Create array of Tau's to get error at (out to 25 days)
Tau_arr = np.arange(25)
    
meanErr_fcst_all = np.full([len(tau0_arr),nDat,len(Tau_arr)], np.nan)

#Get actual hindcast error at each tau0
for iT0 in range(len(tau0_arr)):
    
    #Run LIM script to get appropriate matrices for Tau0
    b_alpha_fnct,L_fnct,Q_fnct,G_fnct,c0_fnct,cT_fnct,u_fnct,v_fnct,g_fnct,periods_fnct,decayT_fnct = LIM(data,tau0_arr[iT0]) #Compute LIM at this lag 
    
    #Make diagonal matrix of g
    g_diag = np.zeros((np.shape(data)[0], np.shape(data)[0]), complex)
    np.fill_diagonal(g_diag, g_fnct)
    
    meanErr_fcst = np.full([nDat, len(Tau_arr)], np.nan) 
    
    #Loop over lead time array (Tau_arr)
    for iTau in range(len(Tau_arr)): 
        # Set up initial values as zero for time-averaged values 
        eps2_L = np.full([nDat],0.)
        ncor_L = np.full([nDat],0.)
        
        #Define Green function at this lag
        if (Tau_arr[iTau] == 0):
            G_tau    = np.dot(u_fnct,np.dot(matrix_power(g_diag,int(Tau_arr[iTau]/tau0_arr[iT0])),np.transpose(v_fnct))).real
        else:
            G_tau    = np.dot(u_fnct,np.dot((g_diag)**(Tau_arr[iTau]/tau0_arr[iT0]),np.transpose(v_fnct))).real
        
        #Define empty array to store forecast at each time
        fcst     = np.full([nDat,nT-Tau_arr[iTau]], np.nan)
        
        for iT in range(nT-Tau_arr[iTau]):    #Loop over timesteps 
        
            # --- Get forecast data --- #
            #If there is any missing data at this time...
            if ( np.isnan(data[:,iT]).any() ): 
                #Indices where NaNs are *not*
                iNan = np.where(~np.isnan(data[:,iT]))

                #Make replica matrix replacing NaNs with zeros so that it won't count for other stations (but is computed)
                replZeros          = np.zeros(np.shape(data[:,iT]))
                replZeros[iNan[0]] = data[iNan[0],iT]

                #Get forecast using zeros array so that stations w/o NaN are still computed properly 
                fcst[:,iT] = np.dot(G_tau, replZeros)

                #Replacing stations that had missing data with a NaN (no forecast if no data)
                fcst[np.where(np.isnan(data[:,iT])),iT] = np.nan
            else:     
                #If no data is missing, just compute forecast as normal - G(Tau) * X(:,t)
                fcst[:,iT] = np.dot(G_tau, data[:,iT])

            # -- Build up mean squared error estimates for the LIM forecast at each station -- #
            for iSt in range(nDat):     #Loop over stations 
                # Make sure data exists
                if ( np.isfinite( data[iSt,iT] ) and np.isfinite( data[iSt,iT+Tau_arr[iTau]] ) ):     
                    fcst_Err    = data[iSt,iT+Tau_arr[iTau]] - fcst[iSt,iT]
                    eps2_L[iSt] = np.nansum([fcst_Err**2, eps2_L[iSt]])
                    ncor_L[iSt] = ncor_L[iSt] + 1

        # Compute mean squared error for every station over the period at this lag 
        meanErr_fcst[:,iTau]         = eps2_L/ncor_L 

    # Set first error estimate to zero: 
    meanErr_fcst[:,0]         = 0

    #Normalize by the variance at each station 
    for iSt in range(nDat):
        meanErr_fcst[iSt,:]   = meanErr_fcst[iSt,:]/np.nanvar(data[iSt,:])

    meanErr_fcst_all[iT0,:,:] = meanErr_fcst  #Should we take the trace of this array? 
        
    print('Finished computing error for Lag = ', tau0_arr[iT0])

#---------------------------------------------------------------------------------------
# Determine tau0 with smallest forecast error over 7-14 day forecasts and all stations
#---------------------------------------------------------------------------------------

meanErr_arr = np.full([len(tau0_arr)], np.nan)

# Find mean error tau0 at 7-14 days lead times and averaged over all soil moisture stations
for iT0 in range(len(tau0_arr)):
    meanErr_arr[iT0] = np.nanmean(meanErr_fcst_all[iT0,int(nDat/2):nDat-1,7:14])

# Test if Q eigenvalues are all positive for tau0 with smallest error
iSort = meanErr_arr.argsort()
for i in iSort:
    Q_test = test_q_eigenvalues(tau0_arr[i],data)
    if(Q_test == True):
        tau0_final = tau0_arr[i]
        print("Lag = ",tau0_final,"days has smallest error")
        break

if(Q_test == False):
    print("Q eigenvalue test failed for all lags")
    print("Do not write LIM parameters and exit program")
    sys.exit()

#-------------------------------------------------------------------------------------------------
# Output best lead (tau0_final), associated Green functions G(tau,tau0_final), and expected error
#-------------------------------------------------------------------------------------------------

b_alpha_fnct,L_fnct,Q_fnct,G_fnct,c0_fnct,cT_fnct,u_fnct,v_fnct,g_fnct,periods_fnct,decayT_fnct = LIM(data,tau0_final)

# Make diagonal matrix of g
g_diag = np.zeros((np.shape(data)[0], np.shape(data)[0]), complex)
np.fill_diagonal(g_diag, g_fnct)
G_tau_3d = np.zeros((len(Tau_arr),np.shape(data)[0], np.shape(data)[0]))

#Empty array to store station errors in
Exp_StationErr = np.full([len(Tau_arr), nDat], np.nan)  

# Loop over lead time array (Tau_arr)
for iTau in range(len(Tau_arr)):
    if (Tau_arr[iTau] == 0):
        G_tau    = np.dot(u_fnct,np.dot(matrix_power(g_diag,int(Tau_arr[iTau]/tau0_final)),np.transpose(v_fnct))).real
    else:
        G_tau    = np.dot(u_fnct,np.dot((g_diag)**(Tau_arr[iTau]/tau0_final),np.transpose(v_fnct))).real

    G_tau_3d[iTau,:,:] = G_tau

    #Get station expected error variance
    b_dot = np.dot(G_tau, np.dot(c0_fnct, np.transpose(G_tau)))
    if iTau > 0:
        errVar = c0_fnct - b_dot
        for iSt in range(nDat):
            Exp_StationErr[iTau, iSt] = np.sqrt(errVar[iSt,iSt])
        
#Set lag=0 value to zero
Exp_StationErr[0,:] = 0

#Normalize by the variance at each station 
#for iSt in range(nDat):
#    Exp_StationErr[iSt,:] = Exp_StationErr[iSt,:]/np.nanvar(data[iSt,:])
    
# Write tau0_final, G, and expected error to output file
lim_matrix_file = os.path.join(opath,"lim_forecast_matrices.p")
f = open(lim_matrix_file,"wb")
pickle.dump(tau0_final, f)
pickle.dump(G_tau_3d, f)
pickle.dump(Exp_StationErr, f)
f.close()


